import { Module, createRequire } from 'module';
import path from 'path';
import vm from 'vm';
import { nanoid } from 'nanoid';
import { fileURLToPath } from 'url';
import { transform, transformSync } from 'esbuild';

const isInESModuleScope = () => {
    try {
        return module === undefined;
    }
    catch {
        return true;
    }
};
const getCallerDirname = () => {
    const _prepareStackTrace = Error.prepareStackTrace;
    Error.prepareStackTrace = (_err, stack) => stack;
    const stack = new Error().stack.slice(2);
    Error.prepareStackTrace = _prepareStackTrace;
    const fileName = stack[0].getFileName();
    return fileName !== null
        ? path.dirname(fileName.startsWith('file://') ? fileURLToPath(new URL(fileName)) : fileName)
        : null;
};
const getEntryDirname = () => path.dirname(process.argv[1]);

const requireFromString = (code, { dirname, globals } = {}) => {
    var _a, _b, _c;
    const mainModule = isInESModuleScope() ? undefined : require.main;
    const moduleDirname = (_b = (_a = dirname !== null && dirname !== void 0 ? dirname : mainModule === null || mainModule === void 0 ? void 0 : mainModule.path) !== null && _a !== void 0 ? _a : getCallerDirname()) !== null && _b !== void 0 ? _b : getEntryDirname();
    const moduleFilename = path.join(moduleDirname, `${nanoid()}.js`);
    const contextModule = new Module(moduleFilename, mainModule);
    contextModule.filename = moduleFilename;
    contextModule.path = moduleDirname;
    contextModule.paths = (_c = mainModule === null || mainModule === void 0 ? void 0 : mainModule.paths) !== null && _c !== void 0 ? _c : [];
    contextModule.require = createRequire(moduleFilename);
    vm.runInNewContext(code, {
        __dirname: contextModule.path,
        __filename: contextModule.filename,
        exports: contextModule.exports,
        module: contextModule,
        require: contextModule.require,
        ...globals
    });
    contextModule.loaded = true;
    return contextModule.exports;
};

const importFromString = async (code, { dirname, globals, transformOptions } = {}) => {
    var _a;
    if (!isInESModuleScope()) {
        ({ code } = await transform(code, {
            format: 'cjs',
            ...transformOptions
        }));
        return requireFromString(code, { dirname, globals });
    }
    // @ts-expect-error: experimental
    if (vm.Module === undefined) {
        throw new Error('command flag `--experimental-vm-modules` is not enabled');
    }
    if (transformOptions !== undefined) {
        ({ code } = await transform(code, {
            format: 'esm',
            ...transformOptions
        }));
    }
    const moduleDirname = (_a = dirname !== null && dirname !== void 0 ? dirname : getCallerDirname()) !== null && _a !== void 0 ? _a : getEntryDirname();
    const moduleFilename = path.join(moduleDirname, `${nanoid()}.js`);
    const context = vm.createContext({
        __IMPORTS__: {},
        ...globals
    });
    // @ts-expect-error: experimental
    const vmModule = new vm.SourceTextModule(code, {
        identifier: moduleFilename,
        context
    });
    // @ts-expect-error: experimental
    const linker = async (specifier) => {
        const importedModuleFilename = new RegExp(`^[.\\${path.sep}]`).test(specifier)
            ? path.resolve(moduleDirname, specifier)
            : undefined;
        const importedModule = await import(importedModuleFilename !== null && importedModuleFilename !== void 0 ? importedModuleFilename : specifier);
        context.__IMPORTS__[specifier] = importedModule;
        const exportedNames = new Set(Object.getOwnPropertyNames(importedModule));
        const importedModuleContent = `${exportedNames.delete('default') ? `export default __IMPORTS__['${specifier}'].default\n` : ''}export const { ${[...exportedNames].join(', ')} } = __IMPORTS__['${specifier}']`;
        // @ts-expect-error: experimental
        return new vm.SourceTextModule(importedModuleContent, {
            identifier: importedModuleFilename !== null && importedModuleFilename !== void 0 ? importedModuleFilename : specifier,
            context
        });
    };
    await vmModule.link(linker);
    await vmModule.evaluate();
    return vmModule.namespace;
};
const importFromStringSync = (code, { dirname, globals, transformOptions } = {}) => {
    if (isInESModuleScope()) {
        throw new Error(`function \`importFromStringSync\` can not work in ES module scope
Use asynchronous function \`importFromString\` instead and execute node with \`--experimental-vm-modules\` command flag.`);
    }
    ({ code } = transformSync(code, {
        format: 'cjs',
        ...transformOptions
    }));
    return requireFromString(code, { dirname, globals });
};

export { importFromString, importFromStringSync, requireFromString };
